"""Synchronize with Airbyte."""
from dataclasses import dataclass, field
from typing import Any, Dict, Optional
import requests
import time
from requests.exceptions import RequestException
from faldbt.logger import LOGGER

from fal.telemetry import telemetry


class AirbyteJobState:
    """Possibe Airbyte job states."""

    RUNNING = "running"
    SUCCEEDED = "succeeded"
    CANCELLED = "cancelled"
    PENDING = "pending"
    FAILED = "failed"
    ERROR = "error"
    INCOMPLETE = "incomplete"


@dataclass
class AirbyteClient:
    """Airbyte REST API connector class."""

    host: str
    max_retries: int = 5
    retry_delay: float = 5
    _base_url: str = field(init=False)

    def __post_init__(self):
        """Set variables."""
        self._base_url = self.host + "/api/v1"

    def request(self, endpoint: str, data: Optional[Dict[str, Any]]):
        """Make a request to Airbyte REST API endpoint."""
        headers = {"accept": "application/json"}

        num_retries = 0
        while True:
            try:
                response = requests.request(
                    method="POST",
                    url=self._base_url + endpoint,
                    headers=headers,
                    json=data,
                    timeout=5,
                )
                response.raise_for_status()
                return response.json()
            except RequestException as e:
                LOGGER.warn(f"Airbyte API request failed: {e}")
                if num_retries == self.max_retries:
                    break
                num_retries += 1
                time.sleep(self.retry_delay)

        raise Exception("Exceeded max number of retries.")

    @telemetry.log_call("airbyte_client_sync")
    def sync(self, connection_id: str) -> dict:
        """Start Airbyte connection sync."""
        return self.request(
            endpoint="/connections/sync", data={"connectionId": connection_id}
        )

    def get_job_status(self, job_id: int) -> dict:
        """Get job status."""
        return self.request(endpoint="/jobs/get", data={"id": job_id})

    def get_connection_data(self, connection_id: str) -> dict:
        """Get details of a connection."""
        return self.request(
            endpoint="/connections/get", data={"connectionId": connection_id}
        )

    @telemetry.log_call("airbyte_client_sync_and_wait")
    def sync_and_wait(
        self,
        connection_id: str,
        interval: float = 10,
        timeout: float = None,
    ):
        """Start a sync operation for the given connector, and polls until done."""
        connection = self.get_connection_data(connection_id)
        job = self.sync(connection_id)
        job_id = job.get("job", {}).get("id")
        LOGGER.info(f"Job {job_id} started for connection: {connection_id}.")
        start = time.monotonic()

        while True:
            if timeout and start + timeout < time.monotonic():
                raise Exception(
                    f"Timeout: Job {job_id} is not finished after {timeout} seconds"
                )

            time.sleep(interval)

            job_details = self.get_job_status(job_id)

            state = job_details.get("job", {}).get("status")

            if state in (
                AirbyteJobState.RUNNING,
                AirbyteJobState.PENDING,
                AirbyteJobState.INCOMPLETE,
            ):
                continue
            elif state == AirbyteJobState.SUCCEEDED:
                break
            elif state == AirbyteJobState.ERROR:
                raise Exception(f"Job failed: {job_id}")
            elif state == AirbyteJobState.CANCELLED:
                raise Exception(f"Job cancelled: {job_id}")
            else:
                raise Exception(f"Unexpected job state `{state}` for job {job_id}")

        return {"job_details": job_details, "connection_details": connection}


@telemetry.log_call("airbyte_sync")
def airbyte_sync(
    host: str,
    connection_id: str,
    interval: float = 10,
    timeout: float = None,
    max_retries: int = 10,
):
    """Sync Airbyte connection."""
    api = AirbyteClient(host=host, max_retries=max_retries)

    return api.sync_and_wait(connection_id, interval=interval, timeout=timeout)



from uuid import uuid4
from typing import List, Any
import os
import argparse
import pkg_resources


LEVEL_FLAGS = {}


class FalArgsError(Exception):
    pass


class _LevelFlag:
    levels: List[str]
    default: Any

    def __init__(self, default):
        self.levels = []
        self.default = default


# This is to handle offering the same flag at several parser levels
# e.g. fal --profiles-dir ~/somewhere run --profiles-dir ~/other_place
# we should get profiles_dir=~/other_place (right-most wins)
def _flag_level(name: str, default=None):
    level = uuid4()
    # Store initial _LevelFlag value in case there is none yet
    LEVEL_FLAGS[name] = LEVEL_FLAGS.get(name, _LevelFlag(default))
    level_flag = LEVEL_FLAGS[name]
    # Add current level, these are meant to read in order (right-most wins)
    level_flag.levels.append(level)

    if default != level_flag.default:
        raise FalArgsError(
            f"Different defaults '{default}' and '{level_flag.default}' for flag '{name}'"
        )

    return f"{name}_{level}"


# Use right after creating the parser, before adding subparsers to it
def _build_fal_common_options(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--disable-logging",
        action="store_true",
        help="Disable logging.",
        dest=_flag_level("disable_logging"),
    )


# Use right after creating the parser, before adding subparsers to it
def _build_dbt_common_options(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--project-dir",
        metavar="PROJECT_DIR",
        help="Directory to look for dbt_project.yml.",
        dest=_flag_level("project_dir", os.getcwd()),
    )

    parser.add_argument(
        "--profiles-dir",
        metavar="PROFILES_DIR",
        help="Directory to look for profiles.yml.",
        dest=_flag_level("profiles_dir"),
    )

    parser.add_argument(
        "--defer",
        action="store_true",
        help="If set, defer to the state variable for resolving unselected nodes.",
        dest=_flag_level("defer"),
    )


def _add_threads_option(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--threads",
        type=int,
        help="Specify number of threads to use while executing Python scripts and dbt models. Overrides settings in profiles.yml.",
    )


def _add_target_option(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--target",
        type=str,
        default=None,
        help="Specify a custom target from profiles.yml.",
    )


def _add_full_refresh_option(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--full-refresh",
        action="store_true",
        default=False,
        help="If specified, fal will pass dbt calls the --full-refresh flag, which will drop incremental models and fully-recalculate the incremental table from the model definition.",
    )


def _add_state_option(parser: argparse.ArgumentParser):
    parser.add_argument("--state", type=str, help="Specify dbt state artifact path")


def _add_vars_option(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--vars",
        type=str,
        default="{}",
        help="""
        Supply variables to the project. This argument overrides variables
        defined in your dbt_project.yml file. This argument should be a YAML
        string, eg. '{my_variable: my_value}'
        """,
    )


def _add_experimental_flow_option(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--experimental-flow",
        action="store_true",
        help="DEPRECATED: no-op",
    )


def _add_experimental_python_models_option(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--experimental-models",
        dest="experimental_python_models",
        action="store_true",
        help="DEPRECATED: no-op",
    )


def _add_experimental_threading(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--experimental-threads",
        type=int,
        help="DEPRECATED: no-op",
        metavar="INT",
    )


def _build_dbt_selectors(sub: argparse.ArgumentParser):

    # fmt: off
    sub.add_argument(
        "-s", "--select",
        nargs="+",
        dest="select",
        help="Specify the nodes to include.",
    )
    sub.add_argument(
        "-m", "--models",
        nargs="+",
        dest="select",
        help="Specify the nodes to include.",
    )
    sub.add_argument(
        "--selector",
        help="The selector name to use, as defined in selectors.yml",
    )
    sub.add_argument(
        "--exclude",
        nargs="+",
        help="Specify the nodes to exclude.",
    )
    # fmt: on


def _build_run_parser(sub: argparse.ArgumentParser):

    # fmt: off
    _build_dbt_selectors(sub)
    _build_dbt_common_options(sub)
    _build_fal_common_options(sub)
    _add_threads_option(sub)
    _add_target_option(sub)

    sub.add_argument(
        "--all",
        action="store_true",
        help="Run scripts for all models. By default, fal runs scripts for models that ran in the last dbt run.",
    )
    sub.add_argument(
        "--scripts",
        nargs="+",
        help="Specify scripts to run, overrides schema.yml",
    )

    sub.add_argument(
        "--before",
        action="store_true",
        help="Run scripts specified in model `before` tag",
    )

    sub.add_argument(
        "--globals",
        action="store_true",
        default=False,
        help="Run global scripts along with selected scripts",
    )
    # fmt: on


def _build_flow_parser(sub: argparse.ArgumentParser):

    flow_command_parsers = sub.add_subparsers(
        title="flow commands",
        dest="flow_command",
        metavar="COMMAND",
        required=True,
    )
    _build_dbt_common_options(sub)
    _build_fal_common_options(sub)

    flow_run_parser = flow_command_parsers.add_parser(
        name="run",
        help="Execute fal and dbt run in correct order",
    )
    _build_dbt_selectors(flow_run_parser)
    _build_dbt_common_options(flow_run_parser)
    _build_fal_common_options(flow_run_parser)
    _add_threads_option(flow_run_parser)
    _add_state_option(flow_run_parser)
    _add_experimental_flow_option(flow_run_parser)
    _add_experimental_python_models_option(flow_run_parser)
    _add_experimental_threading(flow_run_parser)
    _add_vars_option(flow_run_parser)
    _add_target_option(flow_run_parser)
    _add_full_refresh_option(flow_run_parser)


def _build_cli_parser():
    parser = argparse.ArgumentParser(
        prog="fal",
        description="Run Python scripts on dbt models",
    )

    version = pkg_resources.get_distribution("fal").version
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"fal {version}",
        help="show fal version",
    )

    parser.add_argument(
        "--debug",
        action="store_true",
        help="Display debug logging during execution.",
    )

    _build_dbt_common_options(parser)
    _build_fal_common_options(parser)

    # Handle commands
    command_parsers = parser.add_subparsers(
        title="commands",
        dest="command",
        metavar="COMMAND",
        required=True,
    )

    run_parser = command_parsers.add_parser(
        name="run",
        help="Run Python scripts as post-hook nodes",
    )
    _build_run_parser(run_parser)

    flow_parser = command_parsers.add_parser(
        name="flow",
        help="Execute fal and dbt commands in correct order",
    )
    _build_flow_parser(flow_parser)

    return parser


cli_parser = _build_cli_parser()


def parse_args(argv: List[str]) -> argparse.Namespace:
    args = cli_parser.parse_args(argv)
    args_dict = vars(args)

    # Reduce level flags into a single one with the value to use
    for name, level_flag in LEVEL_FLAGS.items():
        args_dict[name] = level_flag.default
        for level in level_flag.levels:
            # Read and delete the level flag to keep only the main one
            current = args_dict.pop(f"{name}_{level}", None)
            if current is not None:
                args_dict[name] = current

    # Build new argparse.Namespace with the correct flags
    return argparse.Namespace(**args_dict)



import os
import sys
import inspect
from pathlib import Path


def create_artifact(context, suffix, additional_data=None):
    model = context.current_model
    model_name = model.name if model else "GLOBAL"
    model_status = model.status if model else None

    output = f"Model name: {model_name}"
    output += f"\nStatus: {model_status}"
    output += f"\nModel dataframe name: {model_name}"
    if additional_data:
        output += f"\n{additional_data}"

    create_file(output, Path(model_name).with_suffix(suffix))


def create_file(output, file_name):
    temp_dir = Path(os.environ["temp_dir"])
    temp_file = temp_dir / file_name
    temp_file.write_text(output)


def create_model_artifact(context, additional_data=None):
    create_artifact(context, ".txt", additional_data)


def create_script_artifact(context, prompt, additional_data=None):
    create_artifact(context, f".{prompt}.txt", additional_data)


def create_dynamic_artifact(context, additional_data=None):
    _, outer_frame, *_ = inspect.stack()
    return create_script_artifact(
        context, Path(outer_frame.filename).stem, additional_data
    )


def get_environment_type():
    # To determine whether this is a fal-created environment or not
    # we'll check whether the executable that is running this script
    # is located under any of the designated fal environment directories.
    from fal.packages.environments.virtual_env import _BASE_VENV_DIR
    from fal.packages.environments.conda import _BASE_CONDA_DIR

    executable_path = Path(sys.executable)
    for environment_type, prefix in [
        ("venv", _BASE_VENV_DIR),
        ("conda", _BASE_CONDA_DIR),
    ]:
        if prefix in executable_path.parents:
            return environment_type
    else:
        return "local"



import argparse
from pathlib import Path
from typing import Any, Dict, List

from dbt.flags import PROFILES_DIR
from fal.planner.executor import parallel_executor
from fal.planner.schedule import Scheduler
from fal.planner.tasks import FalLocalHookTask, Status, TaskGroup

from fal.fal_script import FalScript
from faldbt.project import FAL, DbtModel, FalDbt, FalGeneralException


def create_fal_dbt(
    args: argparse.Namespace, generated_models: Dict[str, Path] = {}
) -> FalDbt:
    profiles_dir = PROFILES_DIR
    if args.profiles_dir is not None:
        profiles_dir = args.profiles_dir

    real_state = None
    if hasattr(args, "state") and args.state is not None:
        real_state = args.state

    return FalDbt(
        args.project_dir,
        profiles_dir,
        args.select,
        args.exclude,
        args.selector,
        args.threads,
        real_state,
        args.target,
        getattr(args, "vars", "{}"),
        generated_models,
    )


def fal_run(args: argparse.Namespace):
    "Runs the fal run command in a subprocess"

    selector_flags = args.select or args.exclude or args.selector
    if args.all and selector_flags:
        raise FalGeneralException(
            "Cannot pass --all flag alongside selection flags (--select/--models, --exclude, --selector)"
        )

    faldbt = create_fal_dbt(args)
    models = _get_filtered_models(faldbt, args.all, selector_flags, args.before)

    scripts = _select_scripts(args, models, faldbt)
    global_scripts = _get_global_scripts(faldbt, args)

    if args.before:
        _handle_global_scripts(args, global_scripts, faldbt, selector_flags)

        pre_hook_scripts = _get_hooks_for_model(models, faldbt, "pre-hook")

        _run_scripts(args, pre_hook_scripts, faldbt)

        _run_scripts(args, scripts, faldbt)

    else:
        _run_scripts(args, scripts, faldbt)

        post_hook_scripts = _get_hooks_for_model(models, faldbt, "post-hook")
        _run_scripts(args, post_hook_scripts, faldbt)
        _handle_global_scripts(args, global_scripts, faldbt, selector_flags)


def _handle_global_scripts(args: argparse.Namespace,
                           global_scripts: List[FalScript],
                           faldbt: FalDbt,
                           selector_flags: Any) -> None:
    scripts_flag = _scripts_flag(args)
    if not scripts_flag and not selector_flags:
        # run globals when no --script is passed and no selector is passed
        _run_scripts(args, global_scripts, faldbt)
    if (scripts_flag or selector_flags) and args.globals:
        _run_scripts(args, global_scripts, faldbt)


def _run_scripts(args: argparse.Namespace, scripts: List[FalScript], faldbt: FalDbt):
    scheduler = Scheduler(
        [TaskGroup(FalLocalHookTask.from_fal_script(script)) for script in scripts]
    )
    parallel_executor(args, faldbt, scheduler)

    failed_tasks: List[FalLocalHookTask] = [
        group.task for group in scheduler.filter_groups(Status.FAILURE)
    ]  # type: ignore
    failed_script_ids = [task.build_fal_script(faldbt).id for task in failed_tasks]
    if failed_script_ids:
        raise RuntimeError(f"Error in scripts {str.join(', ',failed_script_ids)}")


def _scripts_flag(args: argparse.Namespace) -> bool:
    return bool(args.scripts)


def _get_hooks_for_model(
    models: List[DbtModel], faldbt: FalDbt, hook_type: str
) -> List[FalScript]:
    return [
        FalScript.from_hook(faldbt, model, hook)
        for model in models
        for hook in model._get_hooks(hook_type=hook_type)
    ]


def _select_scripts(
    args: argparse.Namespace, models: List[DbtModel], faldbt: FalDbt
) -> List[FalScript]:
    scripts = []
    scripts_flag = _scripts_flag(args)

    for model in models:
        model_scripts = model.get_scripts(before=bool(args.before))
        for path in model_scripts:
            if not scripts_flag:
                # run all scripts when no --script is passed
                scripts.append(FalScript(faldbt, model, path))
            elif path in args.scripts:
                # if --script selector is there only run selected scripts
                scripts.append(FalScript(faldbt, model, path))

    return scripts


def _get_global_scripts(faldbt: FalDbt, args: argparse.Namespace):
    scripts_flag = _scripts_flag(args)
    return [
        FalScript(faldbt, None, path)
        for path in faldbt._global_script_paths["before" if args.before else "after"]
        if not scripts_flag or path in args.scripts
    ]


def _get_models_with_keyword(faldbt: FalDbt) -> List[DbtModel]:
    return list(
        filter(lambda model: FAL in model.meta, faldbt.list_models())
    )


def _get_filtered_models(faldbt: FalDbt, all, selected, before) -> List[DbtModel]:
    selected_ids = _models_ids(faldbt._compile_task._flattened_nodes)
    filtered_models: List[DbtModel] = []

    if (
        not all
        and not selected
        and not before
        and faldbt._run_results.nativeRunResult is None
    ):
        from faldbt.parse import FalParseError

        raise FalParseError(
            "Cannot define models to run without selection flags or dbt run_results artifact or --before flag"
        )

    models = _get_models_with_keyword(faldbt)

    for node in models:
        if selected:
            if node.unique_id in selected_ids:
                filtered_models.append(node)
        elif before:
            if node.get_scripts(before=before) != []:
                filtered_models.append(node)
        elif all:
            filtered_models.append(node)
        elif node.status != "skipped":
            filtered_models.append(node)

    return filtered_models


def _models_ids(models):
    return list(map(lambda r: r.unique_id, models))



from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path
from typing import Iterator, List

import networkx as nx

from faldbt.project import DbtModel
from fal.node_graph import DbtModelNode, NodeGraph, NodeKind, ScriptNode
from fal.planner.tasks import (
    SUCCESS,
    Task,
    DBTTask,
    FalIsolatedHookTask,
    FalLocalHookTask,
    FalModelTask,
    TaskGroup,
    Status,
    HookType,
)
from fal.utils import DynamicIndexProvider
from fal.fal_script import Hook, LocalHook, IsolatedHook, create_hook


def create_hook_task(
    hook: Hook,
    bound_model: DbtModel,
    hook_type: HookType = HookType.HOOK,
) -> Task:
    local_hook = FalLocalHookTask(
        Path(hook.path),
        bound_model=bound_model,
        arguments=hook.arguments,
        hook_type=hook_type,
    )
    if isinstance(hook, IsolatedHook):
        return FalIsolatedHookTask(
            hook.environment_name,
            local_hook,
        )
    else:
        return local_hook


def create_group(
    node: str | nx.DiGraph, properties: dict, node_graph: NodeGraph
) -> TaskGroup:
    kind = properties["kind"]
    if isinstance(node, nx.DiGraph):
        # When we are processing a subgraph, we need to know
        # the exit point of that graph in order to properly
        # bind the post-hooks. We'll do that by sorting each
        # node and putting the `exit_node` at the end (boolean sorting).
        model_ids = sorted(
            list(node),
            key=lambda node: node == properties["exit_node"],
        )
    else:
        model_ids = [node]

    bound_model_name = model_ids[-1]
    flow_node = node_graph.get_node(bound_model_name)

    if kind is NodeKind.DBT_MODEL:
        task = DBTTask(model_ids=model_ids)
    elif kind is NodeKind.FAL_MODEL:
        assert isinstance(flow_node, DbtModelNode)
        model_script = create_hook(
            {
                "path": str(flow_node.model.python_model),
                "environment": properties.get("environment"),
            }
        )
        model_script_task = create_hook_task(
            model_script,
            flow_node.model,
            hook_type=HookType.MODEL_SCRIPT,
        )
        task = FalModelTask(model_ids=model_ids, script=model_script_task)
    else:
        assert kind is NodeKind.FAL_SCRIPT
        assert isinstance(flow_node, ScriptNode)
        task = FalLocalHookTask.from_fal_script(flow_node.script)

    pre_hooks = properties.get("pre_hook", [])
    post_hooks = properties.get("post_hook", [])
    if pre_hooks or post_hooks:
        assert flow_node, "hook nodes must be attached to a model node"
        assert isinstance(flow_node, DbtModelNode)

    pre_hook_tasks = [
        create_hook_task(hook=pre_hook, bound_model=flow_node.model)
        for pre_hook in pre_hooks
    ]
    post_hook_tasks = [
        create_hook_task(hook=post_hook, bound_model=flow_node.model)
        for post_hook in post_hooks
    ]

    return TaskGroup(task, pre_hooks=pre_hook_tasks, post_hooks=post_hook_tasks)


@dataclass
class Scheduler:
    groups: List[TaskGroup]

    def __post_init__(self) -> None:
        index_provider = DynamicIndexProvider()
        for group in self.groups:
            for task in group.iter_tasks():
                task.set_run_index(index_provider)

    def filter_groups(self, status: Status) -> List[TaskGroup]:
        return [group for group in self.groups if group.status is status]

    @property
    def pending_groups(self) -> List[TaskGroup]:
        return self.filter_groups(Status.PENDING)

    def __bool__(self) -> bool:
        return bool(self.pending_groups)

    def _calculate_score(self, target_group: TaskGroup) -> tuple[int, int]:
        # Determine the priority of the group by doing a bunch
        # of calculations. This doesn't really need to be 100% precise,
        # since if we don't have this we'll have to schedule randomly.

        # 1. Number of groups which are only waiting this group (direct dependants)
        # 2. Number of groups which are waiting this group and other groups (indirect dependants)
        # ...

        direct_dependants = 0
        indirect_dependants = 0

        for group in self.pending_groups:
            if group is target_group:
                continue

            if any(dependency is target_group for dependency in group.dependencies):
                indirect_dependants += 1
                if len(group.dependencies) == 1:
                    direct_dependants += 1

        return (direct_dependants, indirect_dependants)

    def _stage_group(self, target_group: TaskGroup) -> None:
        target_group.status = Status.RUNNING

    def finish(self, target_group: TaskGroup, status: int) -> None:
        # When a staged group's execution is finished, we'll remove it
        # altogether and unblock all of its dependencies.

        if status == SUCCESS:
            self._succeed(target_group)
        else:
            self._fail(target_group)

    def _fail(self, target_group: TaskGroup) -> None:
        target_group.status = Status.FAILURE
        for group in self.pending_groups:
            if target_group in group.dependencies:
                group.status = Status.SKIPPED

    def _succeed(self, target_group: TaskGroup) -> None:
        target_group.status = Status.SUCCESS
        for group in self.pending_groups.copy():
            if target_group in group.dependencies:
                group.dependencies.remove(target_group)

    def iter_available_groups(self) -> Iterator[TaskGroup]:
        # Whenever a group is finished we'll remove that from other
        # groups' dependencies. So in here we'll find all unblocked
        # groups (groups without any dependencies) and use the scoring
        # algorithm to determine the priority of each groups (kind of like
        # a dynamic topological sort).
        unblocked_groups = [
            group for group in self.pending_groups if not group.dependencies
        ]
        unblocked_groups.sort(key=self._calculate_score, reverse=True)

        for group in unblocked_groups:
            self._stage_group(group)
            yield group


def schedule_graph(graph: nx.DiGraph, node_graph: NodeGraph) -> Scheduler:
    task_groups = {
        node: create_group(node, properties, node_graph)
        for node, properties in graph.nodes(data=True)
    }

    for name, task_group in task_groups.items():
        task_group.dependencies = [
            task_groups[ancestor] for ancestor in nx.ancestors(graph, name)
        ]

    return Scheduler(list(task_groups.values()))



from __future__ import annotations

from pathlib import Path
from typing import Iterator, List, Optional, Tuple

import importlib_metadata

from fal.utils import cache_static

import importlib_metadata


def _is_fal_pre_release() -> bool:
    from dbt.semver import VersionSpecifier

    raw_fal_version = importlib_metadata.version("fal")
    fal_version = VersionSpecifier.from_version_string(raw_fal_version)
    return fal_version.prerelease


def _get_fal_root_path() -> Path:
    import fal

    # If this is a development version, we'll install
    # the current fal itself.
    base_dir = Path(fal.__file__).parent.parent.parent
    assert (base_dir / ".git").exists()
    return base_dir


def _get_dbt_packages() -> Iterator[Tuple[str]]:
    # package_distributions will return a mapping of top-level package names to a list of distribution names (
    # the PyPI names instead of the import names). An example distirbution info is the following, which
    # contains both the main exporter of the top-level name (dbt-core) as well as all the packages that
    # export anything to that namespace:
    #   {"dbt": ["dbt-core", "dbt-postgres", "dbt-athena-adapter"]}
    #
    # This won't only include dbt.adapters.xxx, but anything that might export anything to the dbt namespace
    # (e.g. a hypothetical plugin that only exports stuff to dbt.includes.xxx) which in theory would allow us
    # to replicate the exact environment.
    package_distributions = importlib_metadata.packages_distributions()
    for dbt_plugin_name in package_distributions.get("dbt", []):
        distribution = importlib_metadata.distribution(dbt_plugin_name)
        if dbt_plugin_name == "dbt-fal" and _is_fal_pre_release():
            adapter_path = _get_fal_root_path() / "adapter"
            yield str(adapter_path), None
        else:
            yield dbt_plugin_name, distribution.version


def _find_fal_extras() -> Iterator[str]:
    # Return a possible set of extras that might be required when installing
    # fal in the new environment. The original form which the user has installed
    # is not present to us (it is not saved anywhere during the package installation
    # process, so there is no way for us to know how a user initially installed fal).
    # We'll therefore be as generous as possible and install all the extras for all
    # the dbt.adapters that the user currently has (so this will still be a subset
    # of dependencies, e.g. if there is no dbt.adapter.duckdb then we won't include
    # duckdb as an extra).

    import pkgutil

    import dbt.adapters

    fal_dist = importlib_metadata.distribution("fal")
    all_extras = fal_dist.metadata.get_all("Provides-Extra", [])

    # This list is different from the one we obtain in _get_dbt_packages
    # since the names here are the actual import names, not the PyPI names
    # (e.g. this one will say athena, and the other one will say dbt-athena-adapter).
    available_dbt_adapters = {
        module_info.name
        for module_info in pkgutil.iter_modules(dbt.adapters.__path__)
        if module_info.ispkg
    }

    # There will be adapters which we won't have an extra for (e.g. oraceledb)
    # and there will be extras which the user did not install the adapter for
    # (e.g. dbt-redshift). We want to take the intersection of all the adapters
    # that the user has installed and all the extras that fal provides and find
    # the smallest possible subset of extras that we can install.
    return available_dbt_adapters.intersection(all_extras)


def _get_fal_package_name() -> Tuple[str, Optional[str]]:
    if _is_fal_pre_release():
        fal_dep, fal_version = str(_get_fal_root_path()), None
    else:
        fal_dep, fal_version = "fal", importlib_metadata.version("fal")

    fal_package_name = fal_dep
    fal_extras = _find_fal_extras()
    if fal_extras:
        fal_package_name += f"[{' ,'.join(fal_extras)}]"
    return fal_package_name, fal_version


def get_default_requirements() -> Iterator[Tuple[str, Optional[str]]]:
    yield _get_fal_package_name()
    yield from _get_dbt_packages()


@cache_static
def get_default_pip_dependencies() -> List[str]:
    return [
        f"{package}=={version}" if version else package
        for package, version in get_default_requirements()
    ]



from dbt.contracts.results import NodeStatus

from faldbt.project import (
    FalDbt,
    DbtModel,
    DbtSource,
    DbtTest,
    DbtGenericTest,
    DbtSingularTest,
)
from fal.fal_script import Context, CurrentModel



import dbt.version
from dbt.semver import VersionSpecifier

def version_compare(version_string: str):
    return DBT_VCURRENT.compare(VersionSpecifier.from_version_string(version_string))

DBT_VCURRENT = dbt.version.installed
IS_DBT_V1PLUS = version_compare("1.0.0") >= 0
IS_DBT_WITH_PYTHON_MODELS = version_compare("1.3.0-a1") >= 0



from dataclasses import dataclass

from dbt import ui
from dbt.logger import log_manager  # For export
from dbt.events.functions import fire_event
from dbt.events.base_types import (
    TestLevel as _TestLevel,
    DebugLevel as _DebugLevel,
    InfoLevel as _InfoLevel,
    WarnLevel as _WarnLevel,
    ErrorLevel as _ErrorLevel,
)

import faldbt.version as version


class FireEventLogger:
    def test(self, msg: str, *args, **kwargs):
        fire_event(TestMessage(_prepare_msg(msg, *args, **kwargs)))

    def trace(self, msg: str, *args, **kwargs):
        # Alias to test
        return self.test(msg, *args, **kwargs)

    def debug(self, msg: str, *args, **kwargs):
        fire_event(DebugMessage(_prepare_msg(msg, *args, **kwargs)))

    def info(self, msg: str, *args, **kwargs):
        fire_event(InfoMessage(_prepare_msg(msg, *args, **kwargs)))

    def warn(self, msg: str, *args, **kwargs):
        fire_event(WarnMessage(_prepare_msg(msg, *args, **kwargs)))

    def warning(self, msg: str, *args, **kwargs):
        # Alias to warn
        return self.warn(msg, *args, **kwargs)

    def error(self, msg: str, *args, **kwargs):
        fire_event(ErrorMessage(_prepare_msg(msg, *args, **kwargs)))


def _prepare_msg(msg: str, *args, **kwargs):
    if args or kwargs:
        return msg.format(*args, **kwargs)
    else:
        return msg


LOGGER = FireEventLogger()

if version.version_compare("1.1.0") < 0:
    from dbt.events.base_types import Cli, Event

    _EXTRA_CLASS_INHERIT = Cli, Event
else:
    _EXTRA_CLASS_INHERIT = (object,)


@dataclass
class TestMessage(_TestLevel, *_EXTRA_CLASS_INHERIT):
    msg: str
    code: str = "FAL0000"

    def message(self) -> str:
        return self.msg


@dataclass
class DebugMessage(_DebugLevel, *_EXTRA_CLASS_INHERIT):
    msg: str
    code: str = "FAL1000"

    def message(self) -> str:
        return self.msg


@dataclass
class InfoMessage(_InfoLevel, *_EXTRA_CLASS_INHERIT):
    msg: str
    code: str = "FAL2000"

    def message(self) -> str:
        return self.msg


@dataclass
class WarnMessage(_WarnLevel, *_EXTRA_CLASS_INHERIT):
    msg: str
    code: str = "FAL3000"

    def message(self) -> str:
        return ui.warning_tag(self.msg)


@dataclass
class ErrorMessage(_ErrorLevel, *_EXTRA_CLASS_INHERIT):
    msg: str
    code: str = "FAL4000"

    def message(self) -> str:
        return self.msg



from __future__ import annotations

import argparse
import threading
import json
from pathlib import Path
import sys
import traceback
import uuid
from functools import partial
from contextlib import contextmanager
from dataclasses import dataclass, field
from enum import Enum, auto
from typing import Iterator, List, Any, Optional, Dict, Tuple, Union

from faldbt.logger import LOGGER

from fal.node_graph import FalScript
from fal.utils import print_run_info, DynamicIndexProvider
from faldbt.project import DbtModel, FalDbt, NodeStatus

from datetime import datetime, timezone

SUCCESS = 0
FAILURE = 1


class Task:
    def set_run_index(self, index_provider: DynamicIndexProvider) -> None:
        self._run_index = index_provider.next()

    @property
    def run_index(self) -> int:
        run_index = getattr(self, "_run_index", None)
        assert run_index is not None
        return run_index

    def execute(self, args: argparse.Namespace, fal_dbt: FalDbt) -> int:
        raise NotImplementedError


class HookType(Enum):
    HOOK = "HOOK"
    SCRIPT = "SCRIPT"
    MODEL_SCRIPT = "MODEL_SCRIPT"


class Status(Enum):
    PENDING = auto()
    RUNNING = auto()
    SKIPPED = auto()
    SUCCESS = auto()
    FAILURE = auto()


def _unique_id_to_model_name(unique_id: str) -> str:
    split_list = unique_id.split(".")
    # if its a unique id 'model.fal_test.model_with_before_scripts'
    return split_list[len(split_list) - 1]


def _unique_ids_to_model_names(id_list: List[str]) -> List[str]:
    return list(map(_unique_id_to_model_name, id_list))


def _mark_dbt_nodes_status_and_response(
    fal_dbt: FalDbt,
    status: NodeStatus,
    dbt_node: Optional[str] = None,
    adapter_response: Optional[dict] = None,
):
    for model in fal_dbt.models:
        if dbt_node is not None:
            if model.unique_id == dbt_node:
                model.status = status

                if adapter_response is not None:
                    model.adapter_response = adapter_response
        else:
            model.status = status


def _map_cli_output_model_results(
    run_results: Dict[Any, Any]
) -> Iterator[Tuple[str, NodeStatus, Optional[dict]]]:
    if not isinstance(run_results.get("results"), list):
        raise Exception("Could not read dbt run results")

    for result in run_results["results"]:
        if not result.get("unique_id") or not result.get("status"):
            continue

        yield result["unique_id"], NodeStatus(result["status"]), result.get(
            "adapter_response"
        )


def _run_script(script: FalScript) -> Dict[str, Any]:
    print_run_info([script])

    # DBT seems to be dealing with only UTC times
    # so we'll follow that convention.
    started_at = datetime.now(tz=timezone.utc)
    try:
        with _modify_path(script.faldbt):
            script.exec()
    except:
        LOGGER.error("Error in script {}:\n{}", script.id, traceback.format_exc())
        # TODO: what else to do?
        status = NodeStatus.Fail
    else:
        status = NodeStatus.Success
    finally:
        LOGGER.debug("Finished script {}", script.id)
        finished_at = datetime.now(tz=timezone.utc)

    return {
        "path": str(script.path),
        "unique_id": str(script.relative_path),
        "status": status,
        "thread_id": threading.current_thread().name,
        "is_hook": script.is_hook,
        "execution_time": (finished_at - started_at).total_seconds(),
        "timing": [
            {
                "name": "execute",
                # DBT's suffix for UTC is Z, but isoformat() uses +00:00. So
                # we'll manually cast it to the proper format.
                # https://stackoverflow.com/a/42777551
                "started_at": started_at.isoformat().replace("+00:00", "Z"),
                "finished_at": finished_at.isoformat().replace("+00:00", "Z"),
            }
        ],
    }


def run_script(script: FalScript, run_index: int) -> int:
    results = _run_script(script)
    run_results_file = Path(script.faldbt.target_path) / f"fal_results_{run_index}.json"
    with open(run_results_file, "w") as stream:
        json.dump(
            {
                "results": [results],
                "elapsed_time": results["execution_time"],
            },
            stream,
        )
    return SUCCESS if results["status"] == NodeStatus.Success else FAILURE


@contextmanager
def _modify_path(fal_dbt: FalDbt):
    sys.path.append(fal_dbt.scripts_dir)
    try:
        yield
    finally:
        sys.path.remove(fal_dbt.scripts_dir)


@dataclass
class DBTTask(Task):
    model_ids: List[str]

    def execute(self, args: argparse.Namespace, fal_dbt: FalDbt) -> int:
        from fal.cli.dbt_runner import dbt_run_through_python

        model_names = _unique_ids_to_model_names(self.model_ids)
        output = dbt_run_through_python(
            args, model_names, fal_dbt.target_path, self.run_index
        )

        for node, status, adapter_response in _map_cli_output_model_results(
            output.run_results
        ):
            _mark_dbt_nodes_status_and_response(fal_dbt, status, node, adapter_response)

        return output.return_code


@dataclass
class FalModelTask(DBTTask):
    script: Union[FalLocalHookTask, FalIsolatedHookTask]

    def set_run_index(self, index_provider: DynamicIndexProvider) -> None:
        super().set_run_index(index_provider)
        self.script.set_run_index(index_provider)

    def execute(self, args: argparse.Namespace, fal_dbt: FalDbt) -> int:
        # Run the ephemeral model
        dbt_result = super().execute(args, fal_dbt)

        # And then run the Python script if it didn't fail.
        if dbt_result != SUCCESS:
            return dbt_result

        script_result = self.script.execute(args, fal_dbt)
        status = NodeStatus.Success if script_result == SUCCESS else NodeStatus.Error
        _mark_dbt_nodes_status_and_response(
            fal_dbt, status, self.script.bound_model.unique_id
        )
        return script_result


@dataclass
class FalLocalHookTask(Task):
    hook_path: Path
    bound_model: Optional[DbtModel] = None
    arguments: Optional[Dict[str, Any]] = None
    hook_type: HookType = HookType.HOOK

    @classmethod
    def from_fal_script(cls, script: FalScript):
        if (
            script.model is not None
            and script.model.python_model is not None
            and script.path == script.model.python_model
        ):
            hook_type = HookType.MODEL_SCRIPT
        else:
            hook_type = HookType.HOOK if script.is_hook else HookType.SCRIPT

        return cls(
            script.path,
            script.model,
            script.hook_arguments,
            hook_type,
        )

    def execute(self, args: argparse.Namespace, fal_dbt: FalDbt) -> int:
        script = self.build_fal_script(fal_dbt)
        return run_script(script, self.run_index)

    def build_fal_script(self, fal_dbt: FalDbt):
        if self.hook_type is HookType.MODEL_SCRIPT:
            return FalScript.model_script(fal_dbt, model=self.bound_model)
        else:
            return FalScript(
                fal_dbt,
                model=self.bound_model,
                path=str(self.hook_path),
                hook_arguments=self.arguments,
                is_hook=self.hook_type is HookType.HOOK,
            )


@dataclass
class FalIsolatedHookTask(Task):
    environment_name: str
    local_hook: FalLocalHookTask

    def set_run_index(self, index_provider: DynamicIndexProvider) -> None:
        super().set_run_index(index_provider)
        self.local_hook.set_run_index(index_provider)

    def execute(self, args: argparse.Namespace, fal_dbt: FalDbt) -> int:
        environment = fal_dbt._load_environment(self.environment_name)
        if environment is None:
            LOGGER.error("Could not find environment: {}", self.environment_name)
            return FAILURE

        with environment.connect() as connection:
            execute_local_task = partial(
                self.local_hook.execute, args=args, fal_dbt=fal_dbt
            )
            result = connection.run(execute_local_task)
            assert isinstance(result, int), result
            return result

    @property
    def bound_model(self) -> DbtModel:
        return self.local_hook.bound_model


@dataclass
class TaskGroup:
    task: Task
    pre_hooks: List[Task] = field(default_factory=list)
    post_hooks: List[Task] = field(default_factory=list)
    dependencies: List[TaskGroup] = field(default_factory=list)
    status: Status = Status.PENDING

    def __post_init__(self):
        self._id = str(uuid.uuid4())

    @property
    def id(self) -> str:
        return self._id

    def iter_tasks(self) -> Iterator[Task]:
        yield from self.pre_hooks
        yield self.task
        yield from self.post_hooks



from dataclasses import dataclass
from enum import Enum
from fal.el.airbyte import AirbyteClient
from fal.el.fivetran import FivetranClient
from typing import Dict


class ELConfigTypes(Enum):
    AIRBYTE = 1
    FIVETRAN = 2


CONNECTION_KEYS = {"AIRBYTE": "connections", "FIVETRAN": "connectors"}


@dataclass
class FalElClient:
    configs: Dict[str, Dict]

    def airbyte_sync(
        self,
        config_name: str,
        connection_id: str = None,
        connection_name: str = None,
        poll_interval: float = 10,
        poll_timeout: float = None,
        max_retries: int = 10,
    ) -> Dict[str, Dict]:
        return self._run_el_sync(
            config_name=config_name,
            connection_id=connection_id,
            connection_name=connection_name,
            poll_interval=poll_interval,
            poll_timeout=poll_timeout,
        )

    def fivetran_sync(
        self,
        config_name: str,
        connector_id: str = None,
        connector_name: str = None,
        poll_interval: float = 10,
        poll_timeout: float = None,
    ) -> Dict[str, Dict]:
        return self._run_el_sync(
            config_name=config_name,
            connection_id=connector_id,
            connection_name=connector_name,
            poll_interval=poll_interval,
            poll_timeout=poll_timeout,
        )

    def _run_el_sync(
        self,
        config_name: str,
        connection_id: str,
        connection_name: str,
        poll_interval: float = 10,
        poll_timeout: float = None,
    ) -> Dict[str, Dict]:

        el_config = self.configs.get(config_name, None)

        if el_config is None:
            raise Exception(f"EL configuration {config_name} is not found.")

        if connection_id is None and connection_name is None:
            raise Exception(
                "Either connection id or connection name have to be provided."
            )

        config_type = el_config.get("type", "").upper()

        if config_type not in [el.name for el in ELConfigTypes]:
            raise Exception(f"EL configuration type {config_type} is not supported.")

        connection_key = CONNECTION_KEYS[config_type]

        if connection_id is None:
            connections = el_config[connection_key]
            connection = next(
                (c for c in connections if c["name"] == connection_name), None
            )
            if connection is None:
                raise Exception(f"Connection {connection_name} not found.")
            connection_id = connection["id"]

        if config_type == ELConfigTypes.AIRBYTE.name:
            client = AirbyteClient(host=el_config["host"])
            return client.sync_and_wait(connection_id)

        elif config_type == ELConfigTypes.FIVETRAN.name:
            client = FivetranClient(
                api_key=el_config["api_key"], api_secret=el_config["api_secret"]
            )
            return client.sync_and_wait(connector_id=connection_id)

        raise Exception("Not implemented")



"""
fal telemetry code uses source code from:

https://github.com/ploomber/ploomber

Modifications are made to ensure that the code works with fal.
"""

import datetime
import http.client as httplib
import warnings
import posthog
import pkg_resources
import yaml
import os
from pathlib import Path
import sys
import uuid
from functools import wraps
from typing import Any, List, Optional
import inspect

import platform


TELEMETRY_VERSION = "0.0.2"
DEFAULT_HOME_DIR = "~/.fal"
CONF_DIR = "stats"
posthog.project_api_key = "phc_Yf1tsGPPb4POvqVjelT3rPPv2c3FH91zYURyyL30Phy"
FAL_HOME_DIR = os.getenv("FAL_HOME_DIR")

# Validate the input of type string
def str_param(item: Any, name: str) -> str:
    if not isinstance(item, str):
        raise TypeError(
            f"TypeError: Variable not supported/wrong type: "
            f"{item}, {name}, should be a str"
        )
    return item


def shutdown():
    posthog.shutdown()
    # HACK: while https://github.com/PostHog/posthog-python/pull/52 happens
    from posthog.request import _session as posthog_session

    posthog_session.close()


def opt_str_param(item: Any, name: str) -> Optional[str]:
    # Can leverage regular string function
    if item is not None and not isinstance(item, str):
        raise TypeError(
            f"TypeError: Variable not supported/wrong type: "
            f"{item}, {name}, should be a str"
        )
    return item


def python_version():
    py_version = sys.version_info
    return f"{py_version.major}.{py_version.minor}.{py_version.micro}"


def is_online():
    """Check if host is online"""
    conn = httplib.HTTPSConnection("www.google.com", timeout=1)

    try:
        conn.request("HEAD", "/")
        return True
    except Exception:
        return False
    finally:
        conn.close()


# Will output if the code is within a container
def is_docker():
    try:
        cgroup = Path("/proc/self/cgroup")
        docker_env = Path("/.dockerenv")
        return (
            docker_env.exists()
            or cgroup.exists()
            and any("docker" in line for line in cgroup.read_text().splitlines())
        )
    except OSError:
        return False


def is_github():
    """Return True if inside a GitHub Action"""
    return os.getenv("GITHUB_ACTIONS") is not None


def is_gitlab():
    """Return True if inside a GitLab CI"""
    return os.getenv("GITLAB_CI") is not None


def get_os():
    """
    The function will output the client platform
    """
    return platform.system()


def dbt_installed_version():
    """Returns: dbt version"""
    try:
        return pkg_resources.get_distribution("dbt-core").version
    except ImportError:
        return


def fal_installed_version():
    """Returns: fal version"""
    try:
        return pkg_resources.get_distribution("fal").version
    except ImportError:
        return


def is_airflow():
    """Returns: True for Airflow env"""
    return "AIRFLOW_CONFIG" in os.environ or "AIRFLOW_HOME" in os.environ


def get_home_dir():
    """
    Checks if fal home was set through the env variable.
    returns the actual home_dir path.
    """
    return FAL_HOME_DIR if FAL_HOME_DIR else DEFAULT_HOME_DIR


def check_dir_exist(input_location=None):
    """
    Checks if a specific directory exists, creates if not.
    In case the user didn't set a custom dir, will turn to the default home
    """
    home_dir = get_home_dir()

    if input_location:
        p = Path(home_dir, input_location)
    else:
        p = Path(home_dir)

    p = p.expanduser()

    if not p.exists():
        p.mkdir(parents=True)

    return p


def check_uid():
    """
    Checks if local user id exists as a uid file, creates if not.
    """
    uid_path = Path(check_dir_exist(CONF_DIR), "uid.yaml")
    conf = read_conf_file(uid_path)  # file already exist due to version check
    if "uid" not in conf.keys():
        uid = str(uuid.uuid4())
        res = write_conf_file(uid_path, {"uid": uid}, error=True)
        if res:
            return f"NO_UID {res}"
        else:
            return uid
    return conf.get("uid", "NO_UID")


def check_stats_enabled():
    """
    Check if the user allows us to use telemetry. In order of precedence:
    1. If FAL_STATS_ENABLED is defined, check its value
    2. If DO_NOT_TRACK is defined, check its value
    3. Otherwise use the value in stats_enabled in the config.yaml file
    """
    if "FAL_STATS_ENABLED" in os.environ:
        return os.environ["FAL_STATS_ENABLED"].lower() == "true"

    if "DO_NOT_TRACK" in os.environ:
        do_not_track = os.environ["DO_NOT_TRACK"].lower()
        return do_not_track == "0" or do_not_track == "false"

    # Check if local config exists
    config_path = Path(check_dir_exist(CONF_DIR), "config.yaml")
    if not config_path.exists():
        write_conf_file(config_path, {"stats_enabled": True})
        return True
    else:  # read and return config
        conf = read_conf_file(config_path)
        return conf.get("stats_enabled", True)


def check_first_time_usage():
    """
    The function checks for first time usage if the conf file exists and the
    uid file doesn't exist.
    """
    config_path = Path(check_dir_exist(CONF_DIR), "config.yaml")
    uid_path = Path(check_dir_exist(CONF_DIR), "uid.yaml")
    uid_conf = read_conf_file(uid_path)
    return config_path.exists() and "uid" not in uid_conf.keys()


def read_conf_file(conf_path):
    try:
        with conf_path.open("r") as file:
            conf = yaml.safe_load(file)
            return conf
    except Exception as e:
        warnings.warn(f"Can't read config file {e}")
        return {}


def write_conf_file(conf_path, to_write, error=None):
    try:  # Create for future runs
        with conf_path.open("w") as file:
            yaml.dump(to_write, file)
    except Exception as e:
        warnings.warn(f"Can't write to config file: {e}")
        if error:
            return e


def _get_telemetry_info():
    """
    The function checks for the local config and uid files, returns the right
    values according to the config file (True/False). In addition it checks
    for first time installation.
    """
    # Check if telemetry is enabled, if not skip, else check for uid
    telemetry_enabled = check_stats_enabled()

    if telemetry_enabled:

        # Check first time install
        is_install = check_first_time_usage()

        # if not uid, create
        uid = check_uid()
        return telemetry_enabled, uid, is_install
    else:
        return False, "", False


def validate_entries(event_id, uid, action, client_time, total_runtime):
    event_id = str_param(str(event_id), "event_id")
    uid = str_param(uid, "uid")
    action = str_param(action, "action")
    client_time = str_param(str(client_time), "client_time")
    total_runtime = opt_str_param(str(total_runtime), "total_runtime")
    return event_id, uid, action, client_time, total_runtime


def log_api(action, client_time=None, total_runtime=None, additional_props=None):
    """
    This function logs through an API call, assigns parameters if missing like
    timestamp, event id and stats information.
    """
    additional_props = additional_props or {}

    event_id = uuid.uuid4()
    if client_time is None:
        client_time = datetime.datetime.now()

    (telemetry_enabled, uid, is_install) = _get_telemetry_info()
    if "NO_UID" in uid:
        additional_props["uid_issue"] = uid
        uid = None

    py_version = python_version()
    docker_container = is_docker()
    gitlab = is_gitlab()
    github = is_github()
    os = get_os()
    online = is_online()

    if telemetry_enabled and online:
        (event_id, uid, action, client_time, total_runtime) = validate_entries(
            event_id, uid, action, client_time, total_runtime
        )
        props = {
            "event_id": event_id,
            "user_id": uid,
            "action": action,
            "client_time": str(client_time),
            "total_runtime": total_runtime,
            "python_version": py_version,
            "fal_version": fal_installed_version(),
            "dbt_version": dbt_installed_version(),
            "docker_container": docker_container,
            "github_action": github,
            "gitlab_ci": gitlab,
            "os": os,
            "telemetry_version": TELEMETRY_VERSION,
            "$geoip_disable": True,  # This disables GeoIp despite the backend setting
            "$ip": None,  # This disables IP tracking
        }

        if is_airflow():
            props["airflow"] = True

        all_props = {**props, **additional_props}

        if "argv" in all_props:
            all_props["argv"] = _clean_args_list(all_props["argv"])

        if is_install:
            posthog.capture(
                distinct_id=uid, event="install_success", properties=all_props
            )

        posthog.capture(distinct_id=uid, event=action, properties=all_props)


# NOTE: should we log differently depending on the error type?
# NOTE: how should we handle chained exceptions?
def log_call(action, args: List[str] = []):
    """Runs a function and logs it"""

    def _log_call(func):
        @wraps(func)
        def wrapper(*func_args, **func_kwargs):

            sig = inspect.signature(func).bind(*func_args, **func_kwargs)
            sig.apply_defaults()
            log_args = dict(map(lambda arg: (arg, sig.arguments.get(arg)), args))
            log_api(
                action=f"{action}_started",
                additional_props={
                    "argv": sys.argv,
                    "args": log_args,
                },
            )

            start = datetime.datetime.now()

            try:
                result = func(*func_args, **func_kwargs)
            except Exception as e:
                log_api(
                    action=f"{action}_error",
                    total_runtime=str(datetime.datetime.now() - start),
                    additional_props={
                        "exception": str(type(e)),
                        "argv": sys.argv,
                        "args": log_args,
                    },
                )
                raise
            else:
                log_api(
                    action=f"{action}_success",
                    total_runtime=str(datetime.datetime.now() - start),
                    additional_props={
                        "argv": sys.argv,
                        "args": log_args,
                    },
                )

            return result

        return wrapper

    return _log_call


def _clean_args_list(args: List[str]) -> List[str]:
    ALLOWLIST = [
        "--disable-logging",
        "--project-dir",
        "--profiles-dir",
        "--defer",
        "--threads",
        "--thread",
        "--state",
        "--full-refresh",
        "-s",
        "--select",
        "-m",
        "--models",
        "--model",
        "--exclude",
        "--selector",
        "--all",
        "--scripts",
        "--script",
        "--before",
        "run",
        "fal",
        "-v",
        "--version",
        "--debug",
        "flow",
        "--vars",
        "--var",
        "--target",
        "--globals"
    ]
    REDACTED = "[REDACTED]"
    output = []
    for item in args:
        if item in ALLOWLIST:
            output.append(item)
        else:
            output.append(REDACTED)
    return output



import ast
from typing import Iterator, List
import astor
import re


def generate_dbt_dependencies(module: ast.Module) -> str:
    """
    Search for dbt function uses and return them as found wrapped in Jinja braces.
    We do not modify them to let dbt decide if they make sense.
    """

    function_calls = _find_function_calls(ast.walk(module))
    ref_calls = _filter_function_calls_by_name(function_calls, "ref")
    source_calls = _filter_function_calls_by_name(function_calls, "source")

    dbt_ast_calls = _filter_constant_calls(ref_calls + source_calls)

    # Convert ast.Calls back to source code
    dbt_function_calls = list(map(astor.to_source, dbt_ast_calls))
    docstring_dbt_functions = _find_docstring_dbt_functions(module)

    lines: List[str] = docstring_dbt_functions + dbt_function_calls

    # Jinja-fy the calls
    return "\n".join(map(lambda s: "{{ " + s.strip() + " }}", lines))


def write_to_model_check(module: ast.Module):
    """
    Make sure a there is a single write_to_model function call on the top level
    """

    all_function_calls = _find_function_calls(ast.walk(module))
    all_wtm_calls = _filter_function_calls_by_name(all_function_calls, "write_to_model")

    assert (
        len(all_wtm_calls) > 0
    ), "There must be at least one write_to_model call in the Python Model"


def _find_function_calls(nodes: Iterator[ast.AST]) -> List[ast.Call]:
    return [node for node in nodes if isinstance(node, ast.Call)]


def _filter_function_calls_by_name(calls: List[ast.Call], func_name: str):
    """
    Analyze all function calls passed to find the ones that call `func_name`.
    """
    return [
        call
        for call in calls
        if isinstance(call.func, ast.Name) and call.func.id == func_name
    ]


def _filter_constant_calls(calls: List[ast.Call]) -> List[ast.Call]:
    """
    Analyze all function calls passed to find the ones with all literal arguments.
    We ignore a `_func(var)` but accept a `_func('model_name')`
    """

    def _is_constant(arg: ast.expr):
        import sys

        if sys.version_info < (3, 8):
            return isinstance(arg, ast.Str)
        else:
            return isinstance(arg, ast.Constant)

    return [call for call in calls if all(map(_is_constant, call.args))]


def _print_node(node: ast.AST):
    """
    For temporary usage during debugging.
    """
    print(
        node,
        *((f, getattr(node, f)) for f in node._fields),
        *((f, getattr(node, f)) for f in node._attributes),
    )


REF_RE = re.compile("ref\\([^)]*\\)")
SOURCE_RE = re.compile("source\\([^)]*\\)")


def _find_docstring_dbt_functions(module: ast.Module) -> List[str]:
    '''
    Simple regex analysis for docstring in top of the file. User can list dependencies one per line, but not multiline.
    Example:

    """
    A Python model with some docstring introduction.

    Dependencies:
    - ref('model')
    - source('some', 'table')
    """
    '''
    docstring = ast.get_docstring(module, True) or ""

    calls = []
    for line in docstring.splitlines():
        calls.extend(REF_RE.findall(line))
        calls.extend(SOURCE_RE.findall(line))

    return calls



import argparse
from typing import List
import sys

from fal.cli.flow_runner import fal_flow_run
import faldbt.lib as lib
from .args import parse_args
from .fal_runner import fal_run
from fal.telemetry import telemetry

import dbt.exceptions
import dbt.ui
from faldbt.logger import log_manager


def cli(argv: List[str] = sys.argv):
    # Wrapper to be able to shutdown telemetry afterwards
    try:
        _cli(argv)
    finally:
        telemetry.shutdown()


@telemetry.log_call("cli")
def _cli(argv: List[str]):
    parsed = parse_args(argv[1:])

    # TODO: do we still need this?
    with log_manager.applicationbound():
        if parsed.debug:
            log_manager.set_debug()

        if parsed.command == "flow":
            if parsed.flow_command == "run":
                _warn_deprecated_flags(parsed)

                exit_code = fal_flow_run(parsed)
                if exit_code:
                    raise SystemExit(exit_code)

        elif parsed.command == "run":
            fal_run(parsed)


# TODO: remove in fal 0.6.0
def _warn_deprecated_flags(parsed: argparse.Namespace):
    if parsed.experimental_flow:
        dbt.exceptions.warn(
            "Flag `--experimental-flow` is DEPRECATED and is treated as a no-op.\n"
            "This flag will make fal error in 0.6"
        )
    if parsed.experimental_python_models:
        dbt.exceptions.warn(
            "Flag `--experimental-models` is DEPRECATED and is treated as a no-op.\n"
            "This flag will make fal error in 0.6"
        )
    if parsed.experimental_threads:
        dbt.exceptions.warn(
            "Flag `--experimental-threads` is DEPRECATED and is treated as a no-op.\n"
            "Using valued passed for `--threads` instead.\n"
            "This flag will make fal error in 0.6"
        )
        # NOTE: take the number of threads to use from the experimental_threads
        if parsed.threads:
            dbt.exceptions.warn(
                f"WARNING: Overwriting `--threads` value with `--experimental-threads` value"
            )
        parsed.threads = parsed.experimental_threads



import os
import sys
import time
import site
from argparse import ArgumentParser

# Isolated processes are really tricky to debug properly
# so we want to a smooth way to enter the process and see
# what is really going on in the case of errors.
#
# For using the debug mode, you first need to set FAL_DEBUG_ISOLATED_RUNNER
# environment variable to "1" on your `fal flow run` command. This will
# make the isolated process hang at the initialization, and make it print
# the instructions to connect to the controller process.
#
# On a separate shell (while letting the `fal flow run` hang), you can
# execute that command to drop into PDB (Python Debugger). With that
# you can observe each step of the connection and run process.
#
#    Starting the process...
#    ============================================================
#
#
#    Debug mode successfully activated. You can start your debugging session with the following command:
#        $ /[...]/fal/venvs/[...]bin/python /[...]/isolated_runner.py --with-pdb A[...]A=
#
#
#    ============================================================
#    Child connection has been established at the bridge b'\x00listener-17368-0'
#    Awaiting the child process to exit at b'\x00listener-17368-0'
#    Isolated process has exited with status: 0
#

IS_DEBUG_MODE = os.getenv("FAL_DEBUG_ISOLATED_RUNNER") == "1"
DEBUG_TIMEOUT = 60 * 15


def run_client(address: str, *, with_pdb: bool = False) -> int:
    from faldbt.logger import LOGGER
    from fal.packages import bridge

    if with_pdb:
        # This condition will only be activated if we want to
        # debug the isolated process by passing the --with-pdb
        # flag when executing the binary.
        breakpoint()

    LOGGER.debug("Trying to create a connection to {}", address)
    with bridge.child_connection(address) as connection:
        LOGGER.debug("Created child connection to {}", address)
        callable = connection.recv()
        LOGGER.debug("Received the callable at {}", address)
        try:
            result = callable()
            exception = None
        except BaseException as exc:
            result = None
            exception = exc
        finally:
            try:
                connection.send((result, exception))
            except BaseException:
                if exception:
                    # If we can't even send it through the connection
                    # still try to dump it to the stdout as the last
                    # resort.
                    import traceback
                    traceback.print_exc(exception)
                raise
        return result


def _get_shell_bootstrap() -> str:
    # Return a string that contains environment variables that
    # might be used during isolated hook's execution.
    return " ".join(
        f"{session_variable}={os.getenv(session_variable)}"
        for session_variable in [
            # PYTHONPATH is customized by the Dual Environment IPC
            # system to make sure that the isolated process can
            # import stuff from the primary environment. Without this
            # the isolated process will not be able to run properly
            # on the newly created debug session.
            "PYTHONPATH",
        ]
        if session_variable in os.environ
    )


def _fal_main() -> None:
    from faldbt.logger import LOGGER
    from fal.packages import bridge

    LOGGER.debug("Starting the isolated process at PID {}", os.getpid())

    parser = ArgumentParser()
    parser.add_argument("listen_at")
    parser.add_argument("--with-pdb", action="store_true", default=False)

    options = parser.parse_args()
    if IS_DEBUG_MODE:
        assert not options.with_pdb, "--with-pdb can't be used in the debug mode"
        message = "=" * 60
        message += "\n" * 3
        message += "Debug mode successfully activated. You can start your debugging session with the following command:\n"
        message += f"    $ {_get_shell_bootstrap()}\\\n     {sys.executable} {os.path.abspath(__file__)} --with-pdb {options.listen_at}"
        message += "\n" * 3
        message += "=" * 60
        LOGGER.info(message)
        time.sleep(DEBUG_TIMEOUT)

    address = bridge.decode_service_address(options.listen_at)
    return run_client(address, with_pdb=options.with_pdb)


def _process_primary_env_packages() -> None:
    python_path = os.getenv("PYTHONPATH")
    if python_path is None:
        return None

    for site_dir in python_path.split(os.pathsep):
        site.addsitedir(site_dir)


def main():
    _process_primary_env_packages()

    from faldbt.logger import log_manager

    # TODO: do we still need this?
    with log_manager.applicationbound():
        _fal_main()


if __name__ == "__main__":
    sys.exit(main())



import multiprocessing
from multiprocessing.connection import Connection
from typing import Any, Dict, Optional, List
import warnings
import json
from faldbt.logger import LOGGER
import os
import argparse


class DbtCliOutput:
    def __init__(
        self,
        command: str,
        return_code: int,
        raw_output: Optional[str],
        logs: Optional[List[Dict[str, Any]]],
        run_results: Dict[str, Any],
    ):
        self._command = command
        self._return_code = return_code
        self._raw_output = raw_output
        self._logs = logs
        self._run_results = run_results

    @property
    def docs_url(self) -> Optional[str]:
        return None

    @property
    def command(self) -> str:
        return self._command

    @property
    def return_code(self) -> int:
        return self._return_code

    @property
    def raw_output(self) -> Optional[str]:
        return self._raw_output

    @property
    def logs(self) -> Optional[List[Dict[str, Any]]]:
        return self._logs

    @property
    def run_results(self) -> Dict[str, Any]:
        return self._run_results


def get_dbt_command_list(args: argparse.Namespace, models_list: List[str]) -> List[str]:
    command_list = []

    if args.debug:
        command_list += ["--debug"]

    command_list += ["run"]

    # NOTE: Safety measure because we do threading on fal
    command_list += ["--threads", str(1)]

    if args.project_dir:
        command_list += ["--project-dir", args.project_dir]
    if args.profiles_dir:
        command_list += ["--profiles-dir", args.profiles_dir]

    if args.defer:
        command_list += ["--defer"]

    if args.state:
        command_list += ["--state", args.state]

    if args.full_refresh:
        command_list += ["--full-refresh"]

    if args.target:
        command_list += ["--target", args.target]

    if args.vars is not None and args.vars != "{}":
        command_list += ["--vars", args.vars]

    if len(models_list) > 0:
        command_list += ["--select"] + models_list

    # Assure all command parts are str
    return list(map(str, command_list))


# This is the Python implementation of the `dbt_run()` function, in which
# we directly use dbt-core as a Python library. We don't run it directly
# but rather use 'multiprocessing' to run it in a real system Process to
# imitate the existing behavior of `dbt_run()` (in terms of performance).


def _dbt_run_through_python(
    args: List[str], target_path: str, run_index: int, connection: Connection
):
    # logbook is currently using deprecated APIs internally, which is causing
    # a crash. We'll mirror the solution from DBT, until it is fixed on
    # upstream.
    #
    # PR from dbt-core: https://github.com/dbt-labs/dbt-core/pull/4866

    warnings.filterwarnings("ignore", category=DeprecationWarning, module="logbook")

    from dbt.main import handle_and_check

    run_results = exc = None
    try:
        run_results, success = handle_and_check(args)
    except BaseException as _exc:
        return_code = getattr(exc, "code", 1)
        exc = _exc
    else:
        return_code = 0 if success else 1

    LOGGER.debug(f"dbt exited with return code {return_code}")

    # The 'run_results' object has a 'write()' method which is basically json.dump().
    # We'll dump it directly to the fal results file (instead of first dumping it to
    # run results and then copying it over).
    if run_results is not None:
        run_results.write(os.path.join(target_path, f"fal_results_{run_index}.json"))
    else:
        connection.send(exc)
        return

    connection.send(return_code)


def dbt_run_through_python(
    args: argparse.Namespace, models_list: List[str], target_path: str, run_index: int
) -> DbtCliOutput:
    """Run DBT from the Python entry point in a subprocess."""
    # dbt-core is currently using the spawn as its mulitprocessing context
    # so we'll mirror it.
    if multiprocessing.get_start_method() != "spawn":
        multiprocessing.set_start_method("spawn", force=True)

    args_list = get_dbt_command_list(args, models_list)

    cmd_str = " ".join(["dbt", *args_list])
    LOGGER.info("Running command: {}", cmd_str)

    # We will be using a multiprocessing.Pipe to communicate
    # from subprocess to main process about the return code
    # as well as the exceptions that might arise.
    p_connection, c_connection = multiprocessing.Pipe()
    process = multiprocessing.Process(
        target=_dbt_run_through_python,
        args=(args_list, target_path, run_index, c_connection),
    )

    process.start()
    result = p_connection.recv()
    if not isinstance(result, int):
        raise RuntimeError("Error running dbt run") from result
    process.join()

    run_results = _get_index_run_results(target_path, run_index)
    return DbtCliOutput(
        command=cmd_str,
        return_code=result,
        raw_output=None,
        logs=None,
        run_results=run_results,
    )


def _get_index_run_results(target_path: str, run_index: int) -> Dict[Any, Any]:
    """Get run results for a given run index."""
    with open(
        os.path.join(target_path, f"fal_results_{run_index}.json")
    ) as raw_results:
        return json.load(raw_results)



from IPython.core.magic import register_line_magic, needs_local_scope
from functools import partial
from fal import FalDbt


@register_line_magic
@needs_local_scope
def init_fal(line="", local_ns={}):
    '''
    Init fal magic variables. Must provide project_dir and profiles_dir.

    Example:
    """
    from faldbt.magics import init_fal

    %init_fal project_dir=/my_project_dir profiles_dir=/my_profiles_dir default_model_name=my_model
    """
    '''
    args = dict([arg.split("=") for arg in line.split()])
    if not args.get("project_dir") or not args.get("profiles_dir"):
        raise Exception(
            """
            Both project_dir and profiles_dir need to be provided:
            Example: %init_fal project_dir=/my_project_dir profiles_dir=/my_profiles_dir
            """
        )

    faldbt = FalDbt(args["project_dir"], args["profiles_dir"])

    fal_globals = {
        "ref": faldbt.ref,
        "source": faldbt.source,
        "write_to_source": faldbt.write_to_source,
        "write_to_firestore": faldbt.write_to_firestore,
        "list_models": faldbt.list_models,
        "list_models_ids": faldbt.list_models_ids,
        "list_sources": faldbt.list_sources,
        "list_features": faldbt.list_features,
        "execute_sql": faldbt.execute_sql,
        "el": faldbt.el,
    }

    if args.get("default_model_name"):
        fal_globals["write_to_model"] = partial(
            faldbt.write_to_model,
            target_1=args.get("default_model_name"),
            target_2=None,
        )

    else:
        fal_globals["write_to_model"] = _raise_no_model_exception

    local_ns.update(fal_globals)


def _raise_no_model_exception():
    raise Exception(
        '''
        Model not found. Please provide a default model name. Example:
        """
        %init_fal project_dir=/my_project_dir profiles_dir=/my_profiles_dir default_model_name=my_model
        """
        '''
    )



# This file is auto-generated by tools/generate_typing_context.py, please
# don't manually alter the contents.

from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import pandas as pd
    from typing import Optional, Dict, List, Protocol, Any
    from faldbt.project import DbtModel, DbtTest, DbtSource, Feature
    from fal.fal_script import Context, CurrentModel

    class _List_Sources(Protocol):
        def __call__(self) -> List[DbtSource]:
            """
            List tables available for `source` usage
            """
            ...

    class _List_Models_Ids(Protocol):
        def __call__(self) -> Dict[str, str]:
            """
            List model ids available for `ref` usage, formatting like `[ref_name, ...]`
            """
            ...

    class _List_Models(Protocol):
        def __call__(self) -> List[DbtModel]:
            """
            List models
            """
            ...

    class _List_Tests(Protocol):
        def __call__(self) -> List[DbtTest]:
            """
            List tests
            """
            ...

    class _List_Features(Protocol):
        def __call__(self) -> List[Feature]:
            ...

    class _Ref(Protocol):
        def __call__(
            self, target_1: str, target_2: Optional[str] = None
        ) -> pd.DataFrame:
            """
            Download a dbt model as a pandas.DataFrame automagically.
            """
            ...

    class _Source(Protocol):
        def __call__(
            self, target_source_name: str, target_table_name: str
        ) -> pd.DataFrame:
            """
            Download a dbt source as a pandas.DataFrame automagically.
            """
            ...

    class _Write_To_Source(Protocol):
        def __call__(
            self,
            data: pd.DataFrame,
            target_source_name: str,
            target_table_name: str,
            *,
            dtype: Any = None,
            mode: str = "append",
        ):
            """
            Write a pandas.DataFrame to a dbt source automagically.
            """
            ...

    class _Write_To_Firestore(Protocol):
        def __call__(self, df: pd.DataFrame, collection: str, key_column: str):
            """
            Write a pandas.DataFrame to a GCP Firestore collection. You must specify the column to use as key.
            """
            ...

    class _Execute_Sql(Protocol):
        def __call__(self, sql: str) -> pd.DataFrame:
            """Execute a sql query."""
            ...

    # Manually introduced annotations, update manually in tools/generate_typing_context.py template.
    class _Write_To_Model(Protocol):
        def __call__(
            self,
            data: pd.DataFrame,
            *,
            dtype: Any = None,
            mode: str = "overwrite",
            target_1: str = ...,
            target_2: Optional[str] = ...,
        ):
            """
            Write a pandas.DataFrame to a dbt model automagically.
            """
            ...


context: Context
write_to_model: _Write_To_Model

list_sources: _List_Sources
list_models_ids: _List_Models_Ids
list_models: _List_Models
list_tests: _List_Tests
list_features: _List_Features
ref: _Ref
source: _Source
write_to_source: _Write_To_Source
write_to_firestore: _Write_To_Firestore
execute_sql: _Execute_Sql



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'tests'}

packages = \
['_fal_testing',
 'fal',
 'fal.cli',
 'fal.cli.model_generator',
 'fal.el',
 'fal.feature_store',
 'fal.packages',
 'fal.packages.environments',
 'fal.planner',
 'fal.telemetry',
 'faldbt',
 'faldbt.utils']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'agate-sql>=0.5.8,<0.6.0',
 'astor>=0.8.1,<0.9.0',
 'backports.functools_lru_cache>=1.6.4,<2.0.0',
 'deprecation>=2.1.0,<3.0.0',
 'dill>=0.3.5.1,<0.4.0.0',
 'firebase-admin>=5.1.0,<6.0.0',
 'google-cloud-bigquery-storage>=2.9.1,<3.0.0',
 'google-cloud-bigquery>=2.28.1,<3.0.0',
 'importlib-metadata>=4.12.0,<5.0.0',
 'pandas>=1.3.4,<2.0.0',
 'platformdirs>=2.5.2,<3.0.0',
 'posthog>=1.4.5,<2.0.0',
 'pyarrow>=6.0.1,<7.0.0',
 'requests>=2.27.1,<3.0.0',
 'virtualenv>=20.16.2,<21.0.0']

extras_require = \
{'dev': ['matplotlib>=3.5.2,<4.0.0'],
 'duckdb': ['duckdb-engine>=0.1.8,<0.2.0'],
 'redshift': ['sqlalchemy-redshift>=0.8.9,<0.9.0'],
 'snowflake': ['snowflake-connector-python==2.7.9'],
 'test': ['pytest>=5.2,<6.0',
          'black>=22.3,<23.0',
          'behave>=1.2.6,<2.0.0',
          'mock>=4.0.3,<5.0.0',
          'pytest-mock>=3.7.0,<4.0.0']}

entry_points = \
{'console_scripts': ['fal = fal.cli:cli']}

setup_kwargs = {
    'name': 'fal',
    'version': '0.7.1',
    'description': 'fal allows you to run python scripts directly from your dbt project.',
    'long_description': '# fal: do more with dbt\n\nfal allows you to run Python scripts directly from your [dbt](https://www.getdbt.com/) project.\n\n- Sign up for the private beta of [fal Cloud](https://fal.ai#signup)\n  \n<p align="center">\n  <a href="https://fal.ai#signup">\n    <img src="https://badgen.net/badge/icon/Sign%20Up%20for%20fal%20Cloud/purple?icon=terminal&label" alt="fal Cloud" />\n  </a>&nbsp;\n  <a href="https://pepy.tech/project/fal">\n    <img src="https://static.pepy.tech/personalized-badge/fal?period=total&units=international_system&left_color=grey&right_color=blue&left_text=Downloads" alt="Total downloads" />\n  </a>&nbsp;\n  <a href="https://pypi.org/project/fal/">\n    <img src="https://badge.fury.io/py/fal.svg" alt="fal on PyPI" />\n  </a>&nbsp;\n  <a href="https://discord.com/invite/Fyc9PwrccF">\n    <img src="https://badgen.net/badge/icon/Join%20Us%20On%20Discord/red?icon=discord&label" alt="Discord conversation" />\n  </a>\n</p>\n\nWith fal, you can:\n\n- Send Slack notifications upon dbt model success or failure.\n- Download dbt models into a Python context with a familiar syntax: `ref(\'my_dbt_model\')`\n- Use Python libraries such as [`sklearn`](https://scikit-learn.org/) or [`prophet`](https://facebook.github.io/prophet/) to build more complex pipelines downstream of **[and in between (new!)](https://blog.fal.ai/python-or-sql-why-not-both/)** `dbt` models.\n\nand more...\n\nCheck out our [Getting Started](#getting-started) guide to get a quickstart, head to our [documentation site](https://docs.fal.ai/) for a deeper dive or play with [in-depth examples](#examples) to see how fal can help you get more done with dbt.\n\n<p align="center">\n  <a href="https://www.loom.com/share/bb49fffaa6f74e90b91d26c77f35ecdc">\n    <img align="center" src="https://cdn.loom.com/sessions/thumbnails/bb49fffaa6f74e90b91d26c77f35ecdc-1637262660876-with-play.jpg" alt="Intro video" />\n  </a>\n</p>\n\n# Getting Started\n\n## 1. Install fal\n\n```bash\n$ pip install fal\n```\n\n## 2. Go to your dbt directory\n\n```bash\n$ cd ~/src/my_dbt_project\n```\n\n## 3. Create a Python script: `send_slack_message.py`\n\n```python\nimport os\nfrom slack_sdk import WebClient\nfrom slack_sdk.errors import SlackApiError\n\nCHANNEL_ID = os.getenv("SLACK_BOT_CHANNEL")\nSLACK_TOKEN = os.getenv("SLACK_BOT_TOKEN")\n\nclient = WebClient(token=SLACK_TOKEN)\nmessage_text = f"Model: {context.current_model.name}. Status: {context.current_model.status}."\n\ntry:\n    response = client.chat_postMessage(\n        channel=CHANNEL_ID,\n        text=message_text\n    )\nexcept SlackApiError as e:\n    assert e.response["error"]\n```\n\n## 4. Add a `meta` section in your `schema.yml`\n\n```yaml\nmodels:\n  - name: historical_ozone_levels\n    description: Ozone levels\n    config:\n      materialized: table\n    columns:\n      - name: ozone_level\n        description: Ozone level\n      - name: ds\n        description: Date\n    meta:\n      fal:\n        scripts:\n          - send_slack_message.py\n```\n\n## 5. (New!) Run `fal flow run` \n\n```bash\n$ fal flow run\n# both your dbt models and fal scripts are run\n```\n\n## 6. Alternatively run `dbt` and `fal` consecutively\n\n```bash\n$ dbt run\n# Your dbt models are run\n\n$ fal run\n# Your python scripts are run\n```\n\n# Examples\n\nTo explore what is possible with fal, take a look at the in-depth examples below. We will be adding more examples here over time:\n\n- [Example 1: Send Slack notifications](examples/slack-example/README.md)\n- [Example 2: Metric forecasting](examples/metric-forecast/README.md)\n- [Example 3: Sentiment analysis on support tickets](examples/sentiment-analysis/README.md)\n- [Example 4: Send event to Datadog](examples/datadog_event/README.md)\n- [Example 5: Incorporate fal in CI/CD workflow](examples/ci_example/README.md)\n- [Example 6: Send data to Firestore](examples/write_to_firestore/README.md)\n- [Example 7: Write dbt artifacts to GCS](examples/write_to_gcs/README.md)\n- [Example 8: Write dbt artifacts to AWS S3](examples/write_to_aws/README.md)\n- [Example 9: Use dbt from a Jupyter Notebook](examples/write_jupyter_notebook/README.md)\n- [Example 10: Read and parse dbt metadata](examples/read_dbt_metadata/README.md)\n- [Example 11: Anomaly Detection](examples/anomaly-detection/README.md)\n\n[Check out the examples directory for more](examples/README.md)\n\n# How it works?\n\n`fal` is a command line tool that can read the state of your `dbt` project and help you run Python scripts after your `dbt run`s by leveraging the [`meta` config](https://docs.getdbt.com/reference/resource-configs/meta).\n\n```yaml\nmodels:\n  - name: historical_ozone_levels\n    ...\n    meta:\n      fal:\n        post-hook:\n          # scripts will run concurrently\n          - send_slack_message.py\n          - another_python_script.py\n```\n\n`fal` also provides useful helpers within the Python context to seamlessly interact with dbt models: `ref("my_dbt_model_name")` will pull a dbt model into your Python script as a [`pandas.DataFrame`](https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.html).\n\n## Running scripts before dbt runs\n\nRun scripts before the model runs by using the `pre-hook:` configuration option.\n\nGiven the following schema.yml:\n\n```\nmodels:\n  - name: boston\n    description: Ozone levels\n    config:\n      materialized: table\n    meta:\n      owner: "@meder"\n      fal:\n        scripts:\n          pre-hook:\n            - fal_scripts/trigger_fivetran.py\n          post-hook:\n            - fal_scripts/slack.py\n```\n\n`fal flow run` will run `fal_scripts/trigger_fivetran.py`, then the `boston` dbt model, and finally `fal_scripts/slack.py`. \nIf a model is selected with a selection flag (e.g. `--select boston`), the hooks associated to the model will always run with it.\n\n```bash\n$ fal flow run --select boston\n```\n\n# Concepts\n\n## profile.yml and Credentials\n\n`fal` integrates with `dbt`\'s `profile.yml` file to access and read data from the data warehouse. Once you setup credentials in your `profile.yml` file for your existing `dbt` workflows anytime you use `ref` or `source` to create a dataframe `fal` authenticates using the credentials specified in the `profile.yml` file.\n\n## `meta` Syntax\n\n```yaml\nmodels:\n  - name: historical_ozone_levels\n    ...\n    meta:\n      owner: "@me"\n      fal:\n        post-hook:\n          - send_slack_message.py\n          - another_python_script.py\n```\n\nUse the `fal` and `post-hook` keys underneath the `meta` config to let `fal` CLI know where to look for the Python scripts. You can pass a list of scripts as shown above to run one or more scripts as a post-hook operation after a `dbt run`.\n\n## Variables and functions\n\nInside a Python script, you get access to some useful variables and functions\n\n### Variables\n\nA `context` object with information relevant to the model through which the script was run. For the [`meta` Syntax](#meta-syntax) example, we would get the following:\n\n```python\ncontext.current_model.name\n#= historical_ozone_levels\n\ncontext.current_model.meta\n#= {\'owner\': \'@me\'}\n\ncontext.current_model.meta.get("owner")\n#= \'@me\'\n\ncontext.current_model.status\n# Could be one of\n#= \'success\'\n#= \'error\'\n#= \'skipped\'\n```\n\n`context` object also has access to test information related to the current model. If the previous dbt command was either `test` or `build`, the `context.current_model.test` property is populated with a list of tests:\n\n```python\ncontext.current_model.tests\n#= [CurrentTest(name=\'not_null\', modelname=\'historical_ozone_levels, column=\'ds\', status=\'Pass\')]\n```\n\n### `ref` and `source` functions\n\nThere are also available some familiar functions from `dbt`\n\n```python\n# Refer to dbt models or sources by name and returns it as `pandas.DataFrame`\nref(\'model_name\')\nsource(\'source_name\', \'table_name\')\n\n# You can use it to get the running model data\nref(context.current_model.name)\n```\n\n### `write_to_model` function\n\nIt is also possible to send data back to your data-warehouse. This makes it easy to get the data, process it and upload it back into dbt territory.\n\nThis function is available in Python Data models only, that is a Python script inside your `models` directory.\nOnce added, it will automatically be run by fal without having to add any extra configurations in the `schema.yml`.\n\n```python\nsource_df = source(\'source_name\', \'table_name\')\nref_df = ref(\'a_model\')\n\n# Your code here\ndf = ...\n\n# Upload a `pandas.DataFrame` back to the datawarehouse\nwrite_to_model(df)\n```\n\n`write_to_model` also accepts an optional `dtype` argument, which lets you specify datatypes of columns. It works the same way as `dtype` argument for [`DataFrame.to_sql` function](https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.to_sql.html).\n\n```python\nfrom sqlalchemy.types import Integer\n# Upload but specifically create the `value` column with type `integer`\n# Can be useful if data has `None` values\nwrite_to_model(df, dtype={\'value\': Integer()})\n```\n\n## Importing `fal` as a Python package\n\nYou may be interested in accessing dbt models and sources easily from a Jupyter Notebook or another Python script.\nFor that, just import the `fal` package and intantiate a FalDbt project:\n\n```py\nfrom fal import FalDbt\nfaldbt = FalDbt(profiles_dir="~/.dbt", project_dir="../my_project")\n\nfaldbt.list_sources()\n# [\n#    DbtSource(name=\'results\' ...),\n#    DbtSource(name=\'ticket_data_sentiment_analysis\' ...)\n#    ...\n# ]\n\nfaldbt.list_models()\n# [\n#    DbtModel(name=\'zendesk_ticket_data\' ...),\n#    DbtModel(name=\'agent_wait_time\' ...)\n#    ...\n# ]\n\n\nsentiments = faldbt.source(\'results\', \'ticket_data_sentiment_analysis\')\n# pandas.DataFrame\ntickets = faldbt.ref(\'stg_zendesk_ticket_data\')\n# pandas.DataFrame\n```\n\n# Supported `dbt` versions\n\nAny extra configuration to work with different `dbt` versions is not needed, latest `fal` version currently supports:\n\n- 1.0.\\*\n- 1.1.\\*\n- 1.2.\\*\n\nIf you need another version, [open an issue](https://github.com/fal-ai/fal/issues/new) and we will take a look!\n\n# Contributing / Development\n\nWe use Poetry for dependency management and easy development testing.\n\nUse Poetry shell to trying your changes right away:\n\n```sh\n~ $ cd fal\n\n~/fal $ poetry install\n\n~/fal $ poetry shell\nSpawning shell within [...]/fal-eFX98vrn-py3.8\n\n~/fal fal-eFX98vrn-py3.8 $ cd ../dbt_project\n\n~/dbt_project fal-eFX98vrn-py3.8 $ fal flow run\n19:27:30  Found 5 models, 0 tests, 0 snapshots, 0 analyses, 165 macros, 0 operations, 0 seed files, 1 source, 0 exposures, 0 metrics\n19:27:30 | Starting fal run for following models and scripts:\n[...]\n```\n\n## Running tests\n\nTests rely on a Postgres database to be present, this can be achieved with docker-compose:\n\n```sh\n~/fal $ docker-compose -f tests/docker-compose.yml up -d\nCreating network "tests_default" with the default driver\nCreating fal_db ... done\n\n# Necessary for the import test\n~/fal $ dbt run --profiles-dir tests/mock/mockProfile --project-dir tests/mock\nRunning with dbt=1.0.1\n[...]\nCompleted successfully\nDone. PASS=5 WARN=0 ERROR=0 SKIP=0 TOTAL=5\n\n~/fal $ pytest -s\n```\n\n# Why are we building this?\n\nWe think `dbt` is great because it empowers data people to get more done with the tools that they are already familiar with.\n\n`dbt`\'s SQL only design is powerful, but if you ever want to get out of SQL-land and connect to external services or get into Python-land for any reason, you will have a hard time. We built `fal` to enable Python workloads (sending alerts to Slack, building predictive models, pushing data to non-data warehose destinations and more) **right within `dbt`**.\n\nThis library will form the basis of our attempt to more comprehensively enable **data science workloads** downstream of dbt. And because having reliable data pipelines is the most important ingredient in building predictive analytics, we are building a library that integrates well with dbt.\n\n# Have feedback or need help?\n\n[Join us in #fal on Discord](https://discord.com/invite/Fyc9PwrccF)\n',
    'author': 'Features & Labels',
    'author_email': 'hello@fal.ai',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/fal-ai/fal',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7.2,<3.11',
}


setup(**setup_kwargs)



"""Synchronize with Fivetran."""
import time
import requests
import datetime
import json
from enum import Enum
from requests.auth import HTTPBasicAuth
from requests.exceptions import RequestException
from typing import Any, Dict, Optional
from urllib.parse import urljoin
from dateutil import parser
from faldbt.logger import LOGGER

BASE_URL = "https://api.fivetran.com/v1/connectors/"


class ScheduleType(Enum):
    """Fivetran connector schedule type."""

    AUTO = 1
    MANUAL = 2


class MethodsEnum(Enum):
    GET = "GET"
    PATCH = "PATCH"
    POST = "POST"
    DELETE = "DELETE"


class FivetranClient:
    """Fivetran API client."""

    def __init__(
        self,
        api_key: str,
        api_secret: str,
        disable_schedule_on_trigger: bool = True,
        max_retries: int = 3,
        retry_delay: float = 0.25,
    ):
        """Initialize with local properties."""
        self.disable_schedule_on_trigger = disable_schedule_on_trigger
        self.max_retries = max_retries
        self.retry_delay = retry_delay
        self._auth = HTTPBasicAuth(api_key, api_secret)

    def _request(
        self,
        *,
        method: MethodsEnum,
        endpoint: Optional[str] = None,
        data: Optional[str] = None,
    ):
        """Make a request to Airbyte REST API endpoint."""
        headers = {"Content-Type": "application/json;version=2"}

        num_retries = 0
        while num_retries <= self.max_retries:
            try:
                response = requests.request(
                    method=method.value,
                    url=urljoin(BASE_URL, endpoint),
                    headers=headers,
                    auth=self._auth,
                    data=data,
                    timeout=5,
                )
                response.raise_for_status()
                parsed = response.json()
                return parsed["data"] if "data" in parsed else parsed
            except RequestException as e:
                LOGGER.warn(f"Fivetran API request failed: {e}")
                num_retries += 1
                time.sleep(self.retry_delay)

        raise Exception("Exceeded max number of retries.")

    def get_connector_data(self, connector_id: str) -> dict:
        """Get details of a connector."""
        return self._request(method=MethodsEnum.GET, endpoint=connector_id)

    def check_connector(self, connector_id: str):
        """Check if connector can be synced."""
        connector_data = self.get_connector_data(connector_id)
        if connector_data["status"]["setup_state"] != "connected":
            raise Exception(f"Cannot sync connector {connector_id}: not set up.")
        if connector_data["paused"]:
            raise Exception(f"Cannot sync connector {connector_id}: paused.")

    def get_sync_status(self, connector_id: str):
        """Return status of the latest sync operation for a given connector."""
        connector_data = self.get_connector_data(connector_id)
        min_time_str = "0001-01-01 00:00:00+00"
        succeeded_at = parser.parse(connector_data["succeeded_at"] or min_time_str)
        failed_at = parser.parse(connector_data["failed_at"] or min_time_str)

        return (
            max(succeeded_at, failed_at),
            succeeded_at > failed_at,
            connector_data["status"]["sync_state"],
        )

    def update_connector(
        self, connector_id: str, properties: Dict[str, Any] = None
    ) -> Dict[str, Any]:
        """Update connector details."""
        return self._request(
            method=MethodsEnum.PATCH, endpoint=connector_id, data=json.dumps(properties)
        )

    def _update_schedule_type(
        self, connector_id: str, schedule_type: ScheduleType = None
    ) -> Dict[str, Any]:
        """Update connector schedule type to "auto" or "manual"."""
        if type(schedule_type) != ScheduleType:
            raise Exception("schedule_type must be either 'auto' or 'manual'.")
        return self.update_connector(
            connector_id, properties={"schedule_type": schedule_type.name.lower()}
        )

    def _get_connector_schema(self, connector_id: str) -> Dict[str, Any]:
        """Get schema config for a connector."""
        return self._request(method=MethodsEnum.GET, endpoint=f"{connector_id}/schemas")

    def start_sync(self, connector_id: str) -> Dict[str, Any]:
        """Start a Fivetran connector sync."""
        if self.disable_schedule_on_trigger:
            LOGGER.info("Disabling Fivetran sync schedule.")
            self._update_schedule_type(connector_id, ScheduleType.MANUAL)
        self.check_connector(connector_id)
        self._request(method=MethodsEnum.POST, endpoint=f"{connector_id}/force")
        connector_data = self.get_connector_data(connector_id)
        LOGGER.info(f"Sync start for connector_id={connector_id}.")
        return connector_data

    def _poll_sync(
        self,
        connector_id: str,
        prev_sync_completion: datetime.datetime,
        poll_interval: float = 10,
        poll_timeout: float = None,
    ) -> Dict[str, Any]:
        """Poll Fivetran until the next sync completes."""
        # The previous sync completion time is necessary because the only way to tell when a sync
        # completes is when this value changes.

        poll_start = datetime.datetime.now()
        while True:
            (
                curr_last_sync_completion,
                curr_last_sync_succeeded,
                curr_sync_state,
            ) = self.get_sync_status(connector_id)
            LOGGER.info(f"Connector '{connector_id}'. State: {curr_sync_state}")

            if curr_last_sync_completion > prev_sync_completion:
                break

            if (
                poll_timeout
                and datetime.datetime.now()
                > poll_start + datetime.timedelta(seconds=poll_timeout)
            ):
                raise Exception(
                    f"Timed out sync for connector '{connector_id}' after {datetime.datetime.now() - poll_start}."
                )

            # Sleep for set poll interval
            time.sleep(poll_interval)

        connector_data = self.get_connector_data(connector_id)
        if not curr_last_sync_succeeded:
            raise Exception("Sync for connector '{connector_id}' failed!")

        return connector_data

    def sync_and_wait(
        self,
        connector_id: str,
        poll_interval: float = 10,
        poll_timeout: float = None,
    ):
        """Start a sync operation for the given connector, and wait until it completes."""
        schema_config = self._get_connector_schema(connector_id)
        init_last_sync_timestamp, _, _ = self.get_sync_status(connector_id)
        self.start_sync(connector_id)
        final_details = self._poll_sync(
            connector_id,
            init_last_sync_timestamp,
            poll_interval=poll_interval,
            poll_timeout=poll_timeout,
        )
        return {"connector_details": final_details, "schema_config": schema_config}



import json
import copy
from pathlib import Path
from typing import Any, Dict, Optional, cast, Union

from fal.cli.fal_runner import create_fal_dbt
from fal.cli.selectors import ExecutionPlan
from fal.cli.model_generator import generate_python_dbt_models
from fal.fal_script import FalScript
from fal.node_graph import DbtModelNode, FalFlowNode, NodeGraph, ScriptNode
from faldbt.project import FalDbt, NodeStatus
import argparse


DBT_RUN_RESULTS_FILENAME = "run_results.json"
FAL_RUN_RESULTS_FILENAME = "fal_results.json"
RUN_RESULTS_KEY = "results"
ELAPSED_TIME_KEY = "elapsed_time"


def run_threaded(
    fal_dbt: FalDbt,
    parsed: argparse.Namespace,
    node_graph: NodeGraph,
) -> int:
    from fal.planner.plan import (
        OriginGraph,
        FilteredGraph,
        PlannedGraph,
        ScriptConnectedGraph,
    )
    from fal.planner.schedule import schedule_graph
    from fal.planner.executor import parallel_executor

    execution_plan = ExecutionPlan.create_plan_from_graph(parsed, node_graph, fal_dbt)

    origin_graph = OriginGraph(node_graph.graph)
    filtered_graph = FilteredGraph.from_execution_plan(
        origin_graph, execution_plan=execution_plan
    )
    connected_graph = ScriptConnectedGraph.from_filtered_graph(filtered_graph)
    planned_graph = PlannedGraph.from_script_connected_graph(
        connected_graph, enable_chunking=False
    )
    scheduler = schedule_graph(planned_graph.graph, node_graph)
    return parallel_executor(parsed, fal_dbt, scheduler)


def fal_flow_run(parsed: argparse.Namespace) -> int:
    arg_vars = getattr(parsed, "vars", "{}")

    # fal-format Python models
    generated_models = generate_python_dbt_models(parsed.project_dir, arg_vars)

    fal_dbt = create_fal_dbt(parsed, generated_models)
    _mark_dbt_nodes_status(fal_dbt, NodeStatus.Skipped)

    node_graph = NodeGraph.from_fal_dbt(fal_dbt)
    exit_code = run_threaded(fal_dbt=fal_dbt, parsed=parsed, node_graph=node_graph)

    # each dbt run creates its own run_results file, here we are combining
    # these files in a single run_results file that fits dbt file format
    _combine_fal_run_results(fal_dbt.target_path)
    return exit_code


def _mark_dbt_nodes_status(
    fal_dbt: FalDbt, status: NodeStatus, dbt_node: Optional[str] = None
):
    for model in fal_dbt.models:
        if dbt_node is not None:
            if model.unique_id == dbt_node:
                model.status = status
        else:
            model.status = status


def node_to_script(node: Union[FalFlowNode, None], fal_dbt: FalDbt) -> FalScript:
    """Convert dbt node into a FalScript."""
    if node is not None and isinstance(node, ScriptNode):
        return cast(ScriptNode, node).script
    elif node is not None and isinstance(node, DbtModelNode):
        return FalScript.model_script(fal_dbt, node.model)
    else:
        raise Exception(f"Cannot convert node to script. Node: {node}")


def _combine_fal_run_results(target_path: str) -> None:
    target_path = Path(target_path)
    dbt_run_results, fal_run_results = [], []
    for path in target_path.glob("fal_results_*.json"):
        assert path.is_file()

        results = _get_all_result_content(path)

        if "dbt_schema_version" in results.get("metadata", {}):
            dbt_run_results.append(results)
        fal_run_results.append(results)

        # Clear out files as we go.
        path.unlink()

    # Use the last DBT result as the framework for putting
    # the rest of the run results.
    if dbt_run_results:
        result_framework = dbt_run_results[-1]
    else:
        result_framework = {
            "metadata": {},
            "args": {},
            ELAPSED_TIME_KEY: float("nan"),
        }

    for file, results in [
        (DBT_RUN_RESULTS_FILENAME, dbt_run_results),
        (FAL_RUN_RESULTS_FILENAME, fal_run_results),
    ]:
        if not results:
            continue

        combined_results = copy.deepcopy(result_framework)
        combined_results[RUN_RESULTS_KEY] = []
        combined_results[ELAPSED_TIME_KEY] = 0.0

        for result in results:
            combined_results[ELAPSED_TIME_KEY] += result.get(ELAPSED_TIME_KEY, 0)
            combined_results[RUN_RESULTS_KEY].extend(result[RUN_RESULTS_KEY])

        with open(target_path / file, "w") as stream:
            json.dump(combined_results, stream)


def _get_all_result_content(file) -> Dict[str, Any]:
    with open(file) as content:
        return json.load(content)
